/*
   2-d Potts model using either the standard local Metropolis algorithm,
   Metropolis, Rosenbluth**2, Teller**2, J. Chem. Phys. 21 (1953) 1087,
   or the cluster updating algorithm of Swendsen and Wang,
   Swendsen and Wang, Phys. Rev. Lett. 58 (1987) 86.
   Parallel version using fast sequential component labeling algorithm on
   each procesor, plus local diffusion to match the labels across 
   processor boundaries. 
   
   For a description of the component labeling, see:
   Baillie and Coddington, Concurrency: Practice and Experience 3, 129 (1991)
   (Caltech preprint C3P-855), or
   Coddington and Baillie, Proc. of the 5th Distributed Memory Computing 
   Conference, Charleston (April 1990), eds. D.W. Walker and Q.F. Stout
   (IEEE Computer Society Press, Los Alamitos, California, 1990),
   (Caltech preprint C3P-862).

   Clive Baillie and Paul Coddington, November 1989.
   
   Revised by Paul Coddington, May 1992.
*/

#include "comext.h"
#include "comdef.h"
#include <stdio.h>
#include <string.h>
#include <math.h>

main()
{
   char ctemp;
   int i,wsweep,swps,iflag, iset;
   int l, anal, initialized;
   FILE *fb,*fd,*faves,*fopen();
   long plop;

   /* Initialize some stuff */

   randx = 0;
   anal = FALSE;
   initialized = FALSE;
   magsq[1] = 0.0;

   powerof2[0] = 1; 
   for (i=1; i<LEVELS; i++) powerof2[i] = 2 * powerof2[i-1]; 

   /* Input parameters */

   printf("Output file format (ASCII or BINARY)? ");
   fflush(stdout);
   scanf("%s", datatype);
   printf("%s\n", datatype);
   fflush(stdout);
   if (strncmp("ASCII",datatype,1) == 0) {
     ASCII = TRUE; BINARY = FALSE;
   }
   else if (strncmp("BINARY",datatype,1) == 0) {
     ASCII = FALSE; BINARY = TRUE;
   }
   else {
     perror("data type not specified");
     exit(-2);
   }

   printf("Q? ");
   fflush(stdout);
   scanf("%d",&Q);
   printf("%d\n",Q);
   fflush(stdout);

   printf("beta? "); 
   fflush(stdout);
   scanf("%f",&beta);
   p = 1.0 - exp(-beta); 
   printf("%12.8f (p=%12.8f)\n",beta,p);
   fflush(stdout);

   for (i=0; i<MAXDIFF+1; i++)
      energy_table[i] = exp( (double) beta * (-i) );

   printf("x,y? ");
   fflush(stdout);
   scanf("%d %d",&globisiz,&globjsiz);
   printf("%d %d\n",globisiz,globjsiz);
   fflush(stdout);
   globsites = globisiz*globjsiz;
   printf("sites: %d\n",globsites);
   volume = (double)globsites;
   fflush(stdout);

   printf("output filename extension? ");
   fflush(stdout);
   scanf("%s",fname) ;
   printf("%s\n",fname) ;
   fflush(stdout);

   printf("first run? ");
   fflush(stdout);
   scanf("%d", &first);
   printf("%d\n",first);
   fflush(stdout);

     /* Create an output file (only used if BINARY) */
     sprintf(outname, "bout.%s", fname);
     if ((fd=fopen(outname,"w")) == NULL) {
        perror("open") ;
	exit( -1 ) ;
     }

   if (BINARY) {
     cfwrite(&Q,sizeof(int),1,fd);
     cfwrite(&beta,sizeof(float),1,fd);
     cfwrite(&isiz,sizeof(int),1,fd);
     cfwrite(&jsiz,sizeof(int),1,fd);
   }

   /* Set up the parallel variables and do domain decomposition
      using the Express gridmap tools, and initialize a bunch
      of things for message passing. */
   decompose();

#ifdef METROPOLIS
   /* Check that detailed balance is OK */
   if ((isiz == 1) && (jsiz == 1)) {
      printf("Only 1 site per processor, so detailed balance is violated!\n");
      fflush(stdout);
      abort(-9);
   }
#endif METROPOLIS

   printf("Enter a menu option (type '?' for a help):\n");
   fflush(stdout);

   /* Read in menu options */
   while ((ctemp=getchar()) != 'Q') {

     switch(ctemp) {

       case 'i' : /* Initialize spins */
		  printf("Start: cold, hot or from file (0,1,2)? ") ;
                  fflush(stdout);
	          scanf("%d",&iflag) ;
	          printf("%d\n",iflag);
                  fflush(stdout);
	          init(iflag) ;
		  if ((iflag == 2) && BINARY) {
		     /* Should write a single value of randx, but it is 
			different for all procs. To save changing the 
			analysis prog, just print out something irrelevant */
		     plop = totswps;
		     cfwrite(&plop,sizeof(long),1,fd);
		  }
		  initialized = TRUE;
	          break;

       case 'a' : /* Analyze data during run */
		  printf("bin size? ");
                  fflush(stdout);
		  scanf("%d", &binsz);
		  printf("%d\n", binsz);
                  fflush(stdout);
		  printf("NA? ");
                  fflush(stdout);
		  scanf("%d", &NA);
		  printf("%d\n", NA);
                  fflush(stdout);
                  anal = TRUE;
		  break;

       case 'u' : /* Update */
		  if (!initialized) {
		    perror("not initialized yet");
		    exit( -3 );
		  }
	          printf("sets? ") ;
                  fflush(stdout);
		  scanf("%d",&sets) ;
		  printf("%d\n",sets) ;
                  fflush(stdout);
		  printf("write sweeps? ") ;
                  fflush(stdout);
		  scanf("%d",&wsweeps) ;
		  printf("%d\n",wsweeps) ;
                  fflush(stdout);
		  printf("sweeps? ") ;
                  fflush(stdout);
		  scanf("%d",&sweeps) ;
		  printf("%d\n",sweeps) ;
                  fflush(stdout);
		  if (BINARY) {
		    cfwrite(&sets,sizeof(int),1,fd);
		    swps = wsweeps * sweeps;
		    cfwrite(&swps,sizeof(int),1,fd);
		  }
		  if (anal) {
		     numbins = (sets * wsweeps) / binsz;
		     analinit(first);
                  }
		  for (iset=0;iset<sets;iset++) {
		     for (wsweep=0;wsweep<wsweeps;wsweep++) {
#ifdef METROPOLIS
		        metupdate(sweeps);
		        measure(spin);
		        metmeasure();
#endif METROPOLIS
#ifdef SW
		        SWupdate(sweeps);
		        measure(spin);
		        SWmeasure();
#endif SW
			if (anal) analdata();
		     }
		     writedata(fd);
		  }
		  totswps += sets*wsweeps*sweeps;
		  break;

       case 'm' : /* Do a measurement */
		  measure(spin);
		  writedata(fd);
		  break;

       case 's':  /* Set the seeds for the random number generator */
		  setseed() ;
		  break;

       case 'b':  /* Back up the spin configuration */
		  sprintf(backname, "back.%s", fname);
                  if ((fb=fopen(backname,"w")) == NULL) {
		     perror("open back") ;
		     exit( -1 ) ;
		  }
		  putdata(fb) ; 
		  fclose(fb) ;
		  break;

       case 'p':  /* Print out the spins */
		  print(spin,0);
		  break;

       case 'l':  /* Print out the cluster labels */
		  printcluster(cluster,0);
		  break;

       case '?' : printf(" i : initialise \n");
		  printf(" u : update # of times \n");
		  printf(" a : analyze data during the run\n");
		  printf(" m : measure energy and total spin \n"); 
          	  printf(" s : set the seed of the rng\n") ;
		  printf(" b : backup the lattice \n");
		  printf(" p : print the lattice \n");
		  printf(" l : print the cluster labels \n");
		  printf(" Q : Quit \n"); 
		  printf(" ? : for this list \n"); 
                  fflush(stdout);
		  break;

     } /* end switch ctemp */

   } /* end while getchar */

   if (anal) {
      /* Print out the averages */
      sprintf(avesname, "aves.%s", fname);
      if ((faves = fopen(avesname, "w")) == NULL) {
         perror("aves open") ;
         exit( -1 ) ;
      }
      printaves(faves);
      fclose(faves);
   }

   /*
   if (ASCII) {
     printf("ns: \n");
     for (i=0;i<sites;i++)
        printf("%d  %d\n", i+1,ns[i]);
     fflush(stdout);
   }
   else if (BINARY) {
     cfwrite(ns,sizeof(int),sites,fd);
   }
   */

   if (BINARY) {
     cfwrite(&totswps,sizeof(int),1,fd);
   }
   printf("totswps: %d\n", totswps);

   /* Print out analyzed data*/
   if (anal) printdata(fname);

   printf("END\n") ;
   fflush(stdout);

#ifdef GRAPHICS
   i = plothwm();
   fmulti(stdout);
   printf("Max bytes used on proc %d was %d for plotix\n", procnum, i);
   fflush(stdout);
   fsingl(stdout);
#endif

   exit( 0 ) ;

} /* end main */

